#include <cv.h>
#include <highgui.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "misc.h"
#include "verbose.h"
#include "timing.h"
#include "video.h"

/* -------------------------------------------------------------------------- */

int threshold2( const CvArr* src, CvArr* dst, double threshold_low,
                double threshold_high, double max_value, int threshold_type )
{
  FUNCTION_BEGIN;
  IplImage *dst_tmp;
  int threshold_type_inv;
  
  switch (threshold_type)
  {
      case CV_THRESH_BINARY:
          threshold_type_inv = CV_THRESH_BINARY_INV;
          break;
      case CV_THRESH_BINARY_INV:
          threshold_type_inv = CV_THRESH_BINARY;
          break;
      case CV_THRESH_TOZERO:
          threshold_type_inv = CV_THRESH_TOZERO_INV;
          break;
      case CV_THRESH_TOZERO_INV:
          threshold_type_inv = CV_THRESH_TOZERO;
          break;
      default:
          PRINT_ERR( "thresholding type not supported" );
          return -1;
  }

  dst_tmp = cvCreateImage( cvGetSize(src), IPL_DEPTH_8U, 1 );
  cvThreshold( src, dst_tmp, threshold_low, max_value, threshold_type_inv );
  cvThreshold( src, dst, threshold_high, max_value, threshold_type );

  switch (threshold_type)
  {
      case CV_THRESH_BINARY:
      case CV_THRESH_TOZERO:
          cvAdd( dst_tmp, dst, dst);
          break;
      default:
          cvAnd( dst_tmp, dst, dst);
  }
  
  cvReleaseImage( &dst_tmp );
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

void greyShading( IplImage* dst )
{
  FUNCTION_BEGIN;
  CvSize imgSize = cvGetSize( dst );
  
  for ( int i = 0; i < imgSize.width; i++ )
      for ( int j = 0; j < imgSize.height; j++ )
          ((uchar*)(dst->imageData + dst->widthStep*j))[i] =
              (char)( i*255./(imgSize.width) );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void redPrediction( IplImage* src, IplImage* dst )
{
  FUNCTION_BEGIN;
  CvSize imgSize = cvGetSize( src );
  uchar *ptr;
  double hue, sat , val, threshold;
  /* Points for two lines decision boundary */
  double line1[2][2] = {{-1,255}, {20,30}};
  double line2[2][2] = {{180-20,30}, {181,255}};
  
/*   if( imgSize != cvGetSize( dst ) ) */
/*   { */
/*     PRINT_ERR( "image sizes are not the same" ); */
/*     return -1; */
/*   } */

  cvSetZero( dst );
  
  for ( int i = 0; i < imgSize.width; i++ )
  {
    for ( int j = 0; j < imgSize.height; j++ )
    {
      ptr = &((uchar*)(src->imageData + src->widthStep*j))[i*3];
      hue = (double)ptr[0];
      sat = (double)ptr[1];
      val = (double)ptr[2];

      if( val < 20 )
          continue;

      /* Prediction unsing two lines decision boundary */
      if( hue > line1[0][0] && hue < line1[1][0]
          && sat > line1[1][1] && sat < line1[0][1] )
      {        
        threshold = (line1[1][1]-line1[0][1]) / (line1[1][0]-line1[0][0])
            * (hue-line1[0][0] ) + line1[0][1];
        if( sat < threshold )
            ((uchar*)(dst->imageData + dst->widthStep*j))[i] = 255;
      }
      else if( hue > line2[0][0] && hue < line2[1][0]
               && sat > line2[0][1] && sat < line2[1][1] )
      {
        threshold = (line2[1][1]-line2[0][1]) / (line2[1][0]-line2[0][0])
            * (hue-line2[0][0] ) + line2[0][1];
        if( sat < threshold )
            ((uchar*)(dst->imageData + dst->widthStep*j))[i] = 255;
      }
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void contourProperties(CvSeq* contour, char* FLAG_ERASE_PTR)
{
  FUNCTION_BEGIN;
  CvSeq *child_contour;
  int i = 0;
  double parent_area = fabs( cvContourArea( contour, CV_WHOLE_SEQ ));
  double child_area;
  double total_area = 0;
  
  PRINT9("parent contour: area = %f", parent_area);
  if( parent_area < 500. || parent_area > 6000.)
      *FLAG_ERASE_PTR = 1;
  
  child_contour = contour->v_next;
  while( child_contour )
  {
    i++;
    child_area = fabs( cvContourArea( child_contour, CV_WHOLE_SEQ ));
    total_area += child_area;
    PRINT9("child contour %d: area = %f", i, child_area );
    child_contour = child_contour->h_next;
  }

  total_area = parent_area - total_area;
  PRINT9("total area = %f", total_area);
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void connectedComponents( IplImage* src, IplImage* dst )
{
  FUNCTION_BEGIN;
  int i = 0;
  char FLAG_ERASE = 0;
  CvScalar black = CV_RGB( 0, 0, 0 );
  CvScalar color = CV_RGB( 255, 255, 255 );
  CvMemStorage *contour_storage = cvCreateMemStorage(0);
  CvSeq *contour = 0;

  cvThreshold( src, src, 1, 255, CV_THRESH_BINARY );
  cvNamedWindow( "Thresholded Frame", 1 );
  cvShowImage( "Thresholded Frame", src );
  cvFindContours( src, contour_storage, &contour, sizeof(CvContour),
                  CV_RETR_CCOMP, CV_CHAIN_APPROX_SIMPLE );

  cvZero( dst );
    
  while(contour)
  {
    i++;
    //CvScalar color = CV_RGB( rand()&255, rand()&255, rand()&255 );
    /* replace CV_FILLED with 1 to see the outlines */
    cvDrawContours( dst, contour, color, color, -1, CV_FILLED, 8 );
    PRINT9("Blob %d", i);
    contourProperties(contour, &FLAG_ERASE);
    if ( FLAG_ERASE )
    {
      cvDrawContours( dst, contour, black, black, -1, CV_FILLED, 8 );
      FLAG_ERASE = !FLAG_ERASE;
    }
    contour = contour->h_next;
  }

  cvNamedWindow( "Surviving Blobs", 1 );
  cvShowImage( "Surviving Blobs", dst );
  cvReleaseMemStorage(&contour_storage);
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void splitPath(const char* path, char** parent, char** name, char** ext)
{
  FUNCTION_BEGIN;
  
  char *pathBegin, *pathEnd, *pname, *pext;
  
  if( path != NULL )
  {
    pathBegin = (char*) path;
    pathEnd = pathBegin + strlen(path);
    pname = strrchr( path, '/' );
    pext = strrchr( path, '.' );
     
    if( pname == NULL )
        pname = pathBegin;
    else
        pname++;

    if( pext <= pname +1 || pext == NULL )
        pext = pathEnd;
  }
  else
  {
    pathBegin = 0;
    pathEnd = 0;
    pname = 0;
    pext = 0;
  }

  if( ext != NULL )
      asprintf( ext, "%.*s",
                (unsigned int) pathEnd - (unsigned int) pext, pext );
  if( name != NULL )
      asprintf( name, "%.*s",
                (unsigned int) pext - (unsigned int) pname, pname );
  if( parent != NULL )
      asprintf( parent, "%.*s",
                (unsigned int) pname - (unsigned int) pathBegin, pathBegin );
  
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

int covEllipse( const CvMat* M, int x1, int x2, CvScalar* ellipse )
{
  FUNCTION_BEGIN;
  double a  = cvmGet( M, x1, x1 );
  double c1 = cvmGet( M, x1, x2 );
  double c2 = cvmGet( M, x2, x1 );
  double b  = cvmGet( M, x2, x2 );
  
  if( CV_MAT_TYPE( M->type ) != CV_32FC1) // To do: use cvGetElemType to test element type.
  {
    PRINT_ERR( "only float element in CvMat is implemented" );
    FUNCTION_END;
    return -1;
  }
    
  if( a < 0 || b < 0)
  {
    PRINT_ERR( "variance is negative! v1 = %f, v2 = %f", a, b );
    FUNCTION_END;
    return -1;
  }
  
  if( c1 != c2 )
  {
    PRINT_ERR( "covariance matrix is not symmetric! %f != %f", c1, c2 );
    //FUNCTION_END;
    //return -1;
  }

  if( c1*c1 > a*b )
  {
    fprintf( stderr, "covEllipse: bad correlation coefficient! rho = %f\n",
             c1 / sqrt( a*b ) );
    return -1;
  }

  /* Ellipse parameters */
  //static double t = 0, i = 1, imax = 1e7; CHRONO(); // for chrono
#define EIG_CLOSED_FORM_2
#ifdef PARAM_CLOSED_FORM
  /* Method: parameters in closed-form (1.423515 us on 800 MHz
   * Duron) */ 
  if( fabs( c1 ) < 1000 * DBL_EPSILON )
  {
    ellipse->val[2] = 0;
    ellipse->val[0] = sqrt( a );
    ellipse->val[1] = sqrt( b );
  }
  else
  {    
    ellipse->val[2] = atan2(2*c1, (a-b)) / 2 ;
    double tmp = c1 * tan( ellipse->val[2] );
    ellipse->val[0] = sqrt( fabs( a + tmp) ) ;
    ellipse->val[1] = sqrt( fabs( b - tmp) ) ;
  }
#endif
#ifdef EIG_CLOSED_FORM_1
  /* Method: eigenvalues in closed-form (1.259138 us on 800 MHz
   * Duron) */
  double tmp = norm2( a-b, 2*c1 );
  
  /* Eigenvalues */
  double lam1 = 0.5 * (a + b + tmp);
  double lam2 = 0.5 * (a + b - tmp);

  /* Ellipse parameters */
  if( fabs( b-lam1 ) < 1000 * DBL_EPSILON )
      ellipse->val[2] = M_PI_2;
  else
      ellipse->val[2] = atan( sqrt( (a-lam1) / (b-lam1) ));
  if( c1 < 0) ellipse->val[2] = - ellipse->val[2];
  ellipse->val[0] = sqrt( lam1 );
  ellipse->val[1] = sqrt( lam2 );
#endif
#ifdef EIG_CLOSED_FORM_2
  /* Method: eigenvalues in closed-form 2 (the fastest, 1.059574 us on
   * 800 MHz Duron) */
  double tmp = norm2( a-b, 2*c1 );
  if (a < b)
      tmp = -tmp;
    
  /* Eigenvalues */
  double lam1 = 0.5 * (a + b + tmp);
  double lam2 = 0.5 * (a + b - tmp);

  /* Ellipse parameters */
  if( fabs(c1) < 1000 * DBL_EPSILON )
      ellipse->val[2] = 0;
  else
      ellipse->val[2] = atan( (lam1-a) / c1 );
  ellipse->val[0] = sqrt( lam1 );
  ellipse->val[1] = sqrt( lam2 );
#endif
#ifdef SVD_DECOMPOSITION
  /* Method: SVD (the slowest, 13.167459 us on 800 MHz Duron) */
  CvMat *C = cvCreateMat( 2, 2, CV_64FC1 );
  ((double*)(C->data.ptr + C->step*0))[0] = a;
  ((double*)(C->data.ptr + C->step*0))[1] = c1;
  ((double*)(C->data.ptr + C->step*1))[0] = c2;
  ((double*)(C->data.ptr + C->step*1))[1] = b;

  CvMat *eigval = cvCreateMat( 2, 1, CV_64FC1 );
  CvMat *eigvec = cvCreateMat( 2, 2, CV_64FC1 );
  cvSVD( C, eigval, NULL, eigvec, CV_SVD_MODIFY_A | CV_SVD_V_T );

  /* Ellipse parameters */
  ellipse->val[0] = sqrt( ((double*)(eigval->data.ptr + eigval->step*0))[0] );
  ellipse->val[1] = sqrt( ((double*)(eigval->data.ptr + eigval->step*1))[0] );
  ellipse->val[2] = atan( ((double*)(eigvec->data.ptr + eigvec->step*0))[1]
                          / ((double*)(eigvec->data.ptr + eigvec->step*0))[0] );

  cvReleaseMat( &C );
  cvReleaseMat( &eigval );
  cvReleaseMat( &eigvec );
#endif
  /* for chrono */
  //t += CHRONO();
  //if(i == imax)
  //    PRINT9("method %d, %f tests, total time: %f us ,mean: %f us", method, i, t, t/i);
  //i++;

  ellipse->val[3] = 0;
  PRINT9( "covariance matrix = [%.3e, %.3e, %.3e, %.3e]", a, c1, c2, b );
  PRINT9( "covariance ellipse = [%.3e, %.3e, %.3e rad = %.1f deg, %.3e]",
           ellipse->val[0], ellipse->val[1], ellipse->val[2],
           ellipse->val[2] * 180 / M_PI, ellipse->val[3] );
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

void drawEllipse( CvArr* img, CvPoint center, CvSize axes, double angle,
                   double start_angle, double end_angle, CvScalar color,
                   int thickness, int line_type, int shift )
{
  FUNCTION_BEGIN;
  PRINT9( "center = (%d,%d); axis = %d,%d; angle = %.1f deg",
           center.x, center.y, axes.width, axes.height, angle );

  if( center.x < VIDEO_WIDTH && center.y < VIDEO_HEIGHT
      && center.x >= 0 && center.y >= 0
      && axes.width < VIDEO_WIDTH && axes.height < VIDEO_HEIGHT
      && axes.width >= 0 && axes.height >= 0 )
  {
    cvEllipse( img, center, axes, angle, start_angle, end_angle, color,
               thickness, line_type, shift );
  }
  else
  {
    PRINT_ERR( "ellipse out of range" );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void sortMatrix64f( CvMat* mat, int row )
{
  FUNCTION_BEGIN;
  const int numberOfRows = mat->rows;
  const int numberOfCols = mat->cols;
  const int numberOfChannels = 1;
  int firstRow, lastRow;
  
  if( row < 0 )
  {
    firstRow = 0;
    lastRow = numberOfRows - 1;
  }
  else
  {
    firstRow = row;
    lastRow = firstRow;
  }
  
  int i, j, k;
  double a;
  for( i = firstRow; i <= lastRow; i++ )
  {
    double* line = (double*)(mat->data.ptr + mat->step*i);
    for( j = 1; j < numberOfCols; j++ )
    {
      /* Pick out each element in turn. */
      a = line[j * numberOfChannels];
      k = j - 1;
      int index = k * numberOfChannels;
      while( k >= 0 && line[index] > a )
      {
        /* Look for the place to insert it. */
        line[index + numberOfChannels] = line[index];
        k--;
        index = k * numberOfChannels;
      }
      line[index + numberOfChannels] = a; /* Insert it. */
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void displayImage( const char* window, IplImage* image, CvFont* font,
                   const char* text, CvScalar fgColor, CvScalar bgColor )
{
  FUNCTION_BEGIN;
  cvNamedWindow( window, 1 );
  if( text != NULL && &font != NULL )
  {
    putText( image, text, cvPoint(0, VIDEO_HEIGHT - 6), font, fgColor, bgColor, 1 );
  }
  if( image != NULL )
  {
    cvShowImage( window, image );
  }
  else
  {
    PRINT_ERR("NULL image");
  }
  //cvWaitKey();
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void putText( IplImage* img, const char* text, CvPoint org, const CvFont* font,
              CvScalar fgColor, CvScalar bgColor, int bgAllWidth )
{
  FUNCTION_BEGIN;
  if( text == NULL )
  {
    PRINT_ERR("text is NULL");
    FUNCTION_END;
    return;
  }
  CvSize textSize;
  int baseline;
  CvPoint topLeft;
  CvPoint bottomRight;
  cvGetTextSize( text, font, &textSize, &baseline );
  if( bgAllWidth == 0 )
  {
    topLeft = cvPoint( org.x, org.y - textSize.height - 1 );
    bottomRight = cvPoint( topLeft.x + textSize.width,
                           topLeft.y + textSize.height + baseline - 1 );
  }
  else
  {
    topLeft = cvPoint( 0, org.y - textSize.height - 1 );
    bottomRight = cvPoint( img->width,
                           topLeft.y + textSize.height + baseline - 1 );
  }
  cvRectangle( img, topLeft, bottomRight, bgColor, CV_FILLED );
  cvPutText( img, text, org, font, fgColor );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

int permuteMatrix32f( CvMat *mat, int a, int b, int rowOrCol )
{
  FUNCTION_BEGIN;
  int length, n, step;
  float swap;
  int type = CV_MAT_TYPE(mat->type);

  if( type != CV_32FC1 )
  {
    PRINT_ERR( "matrix element type (%d) not supported", type );
    FUNCTION_END;
    return -1;
  }
  
  if( rowOrCol == 0) // permutes lines
  {
    length = mat->cols;
    n = 1;
    step = mat->step;
  }
  else // permutes columns
  {
    length = mat->rows;
    n = mat->cols;
    step = sizeof(float);
  }
  
  if( a >= length || b >= length )
  {
    PRINT_ERR( "index exceedes matrix dimensions: %d or %d >= %d", a, b, length );
    FUNCTION_END;
    return -1;
  }
  
  char* data = (char*)(mat->data.ptr);
  float* range1 = (float*)(data + step * a);
  float* range2 = (float*)(data + step * b);
  
  for( int i = 0; i < length ; i++  )
  {
    CV_SWAP( range1[i*n], range2[i*n], swap );
  }
  
  return 0;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void doubleProduct32f( CvMat *A, CvMat *B, CvMat *C, CvMat *dest )
{
  FUNCTION_BEGIN;
  CvMat* tmpMatrix = cvCreateMat( A->rows, B->rows, CV_32FC1 );
  cvGEMM( A, B, 1, 0, 1, tmpMatrix, 0 );
  cvGEMM( tmpMatrix, C, 1, 0, 1, dest, CV_GEMM_B_T );
  cvReleaseMat( &tmpMatrix );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

int cvtRedNorm( const IplImage* src, IplImage* dst )
{
  FUNCTION_BEGIN;
  CvSize srcSize = cvGetSize(src);
  CvSize dstSize = cvGetSize(dst);

  if( dstSize.width != srcSize.width || dstSize.height != srcSize.height )
  {
    PRINT_ERR( "images size differ" );
    FUNCTION_END;
    return -1;
  }

  if( src->nChannels != 3 )
  {
    PRINT_ERR( "source image has %d channel(s) but 3 is required", src->nChannels );
    FUNCTION_END;
    return -1;
  }

  if( dst->nChannels != 1 )
  {
    PRINT_ERR( "source image has %d channel(s) but 1 is required", dst->nChannels );
    FUNCTION_END;
    return -1;
  }
  
  for ( int i = 0; i < srcSize.width; i++ )
      for ( int j = 0; j < srcSize.height; j++ )
      {
        uchar* srcPixel = &((uchar*)(src->imageData + src->widthStep*j))[i*3];
        uchar* dstPixel = &((uchar*)(dst->imageData + dst->widthStep*j))[i];
        *dstPixel = redNorm( srcPixel );
      }
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

void saveOptions( const char* fileName )
{
  FUNCTION_BEGIN;
  char text[512];
  sprintf( text,
           "SIGNS_MAX %d\n"
           "DISTANCE_MAX %f\n"
           "DETECTIONS_MAX %d\n"
           "MISSING_MEASURES_MAX %d\n"
           "AGE_TO_BE_A_SIGN %d\n"
           "HAAR_CASCADE_PATH %s\n"
           "HAAR_SCALE %f\n"
           "HAAR_NEIGHBORS %d\n",
           SIGNS_MAX,
           DISTANCE_MAX,
           DETECTIONS_MAX,
           MISSING_MEASURES_MAX,
           AGE_TO_BE_A_SIGN,
           HAAR_CASCADE_PATH,
           HAAR_SCALE,
           HAAR_NEIGHBORS
           );
  FILE* fd = fopen( fileName, "w" );
  fputs( text, fd );
  fclose(fd);
  FUNCTION_END;
}
