#ifndef _MISC_H
#define _MISC_H

#include <cv.h>
#include <highgui.h>
#include <math.h>
#include "color.h"

/* Return index in a circular buffer. */
#define INDEX_CIRC_BUFFER(index, bufferSize) \
((((index)%(bufferSize))+(bufferSize))%(bufferSize))

/* -------------------------------------------------------------------------- */

/** Return L2-norm: (a^2+b^2)^0.5 */
inline double norm2( double a, double b)
{
  return sqrt( a*a + b*b );
}

/** Return Euclidian distance between two points. */
inline double dist2( CvPoint2D64f a, CvPoint2D64f b)
{
  return norm2( a.x - b.x, a.y - b.y );
}

/** Return normalized red, i.e., Rnorm = 255 * R / (R+G+B). */
inline int redNorm( uchar* BGRPixel )
{
  if( BGRPixel[2] != 0 )
      return (uchar) round( 255. * BGRPixel[2]
                            / (BGRPixel[0] + BGRPixel[1] + BGRPixel[2]) );
  else
      return 0;
}

/* Double thresholding working like cvThreshold function. Values
 * within interval (threshold_low, threshold_high] are processed like
 * values less or equal to threshold in cvThreshold function. */
int threshold2( const CvArr* src, CvArr* dst, double threshold_low,
                double threshold_high, double max_value, int threshold_type );

/* Fill a image with horizontal grey shading. Useful for some tests.
 * dst is a 8-bit 1-channel image. */
void greyShading( IplImage* dst );

/* Predict if a pixel belongs to a red sign. src is a 8-bit 3-channel
 * image. dst is a 8-bit 1-channel image (destination mask). */
void redPrediction(IplImage* src, IplImage* dst );

/* Compute contour properties and check some requirements for a blob
 * to be a sign. */
void contourProperties(CvSeq* contour, char* FLAG_ERASE_PTR);

/* Find connected components and allows delete those which haven't
 * requirements wanted by contourProperties. */
void connectedComponents( IplImage* src, IplImage* dst );

/* Split a path into its components. e.g. /home/foo/file.ext is
 * splitted into /home/foo/, file, .ext. Don't forget to free allocated
 * memory pointed by parent, name, ext. */
void splitPath( const char *path, char **parent, char **name, char **ext );
    
/** Extract error ellipse parameters from covariance matrix.
 *
 * Compute lengths of the half first axis a, half second axis b and
 * orientation angle (from horizontal axis to first axis) of the
 * covaraice ellipse (one stantard deviation unit) given the
 * covariance matrix [M(x1,x1) M(x1,x2); M(x1,x2) M(x2,x2)] in Matlab
 * notation. The function writes ellipse=[a, b, angle(rad), 0].
 */
int covEllipse( const CvMat* M, int x1, int x2, CvScalar* ellipse );
    
/* Works exactly as cvEllipse but can displays things. */
void drawEllipse( CvArr* img, CvPoint center, CvSize axes, double angle,
                  double start_angle, double end_angle, CvScalar color,
                  int thickness = 1, int line_type = 8, int shift = 0 );

/* If row < 0, sorts each line of a matrix into ascending numerical
 * order. If 0 <= row <= mat->rows, sorts row-th line. Sorting is made
 * by straight insertion. Complexity is O(n^2), algorithm is fast if
 * number of elements is less than 20. Otherwise you should use
 * another algorithm like Shell sort or Quick sort. */
void sortMatrix64f( CvMat* mat, int row = -1 );

/* Display image with a bottom text. window is created if doesn't exist. */
void displayImage( const char* window, IplImage* image,
                   CvFont* font = NULL, const char* text = NULL,
                   CvScalar fgColor = CV_BLACK, CvScalar bgColor = CV_WHITE );

/* Same as cvPutText but include a bakground color. If the additionnal
 * parameter bgAllWidth is different of zero, background color is
 * drawn on all width of the image. */
void putText( IplImage* img, const char* text, CvPoint org, const CvFont* font,
              CvScalar fgColor = CV_BLACK, CvScalar bgColor = CV_WHITE,
              int bgAllWidth = 0 );

/* Permute two lines (rowOrCol = 0) or two columns (rowOrCol != 0) of
 * 32 bits floating element matrix mat. */
int permuteMatrix32f( CvMat *mat, int a, int b, int rowOrCol );

/* Compute A*B*C^T and put it into dest. */
void doubleProduct32f( CvMat *A, CvMat *B, CvMat *C, CvMat *dest );

/* Convert a RGB image to normalized red, i.e.,
 * Rnorm = 255 * R / (R+G+B). */
int cvtRedNorm( const IplImage* src, IplImage* dst );

/* Write in fileName some options. */
void saveOptions( const char* fileName );

#endif
