#include <cv.h>
#include <cvaux.h>
#include <stdio.h>
#include "tracking.h"
#include "verbose.h"
#include "misc.h"
#include "color.h"

/* -------------------------------------------------------------------------- */

int initTracking( Tracking* tracking)
{
  FUNCTION_BEGIN;
  tracking->numberOfSigns = 0;
  
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    PRINT9("init sign %d", i);
    Sign* sign = &tracking->signs[i];
    /* Kalman */
    CvKalman* kalman = cvCreateKalman( NUMBER_OF_STATES,
                                       NUMBER_OF_MEASURES,
                                       NUMBER_OF_CONTROLS );
    sign->kalman = kalman;
    CvMat* m = kalman->measurement_matrix;
    cvSetZero(m);
    CV_MAT_ELEM( *m, float, M_X, S_X ) = 1;
    CV_MAT_ELEM( *m, float, M_Y, S_Y ) = 1;
    CV_MAT_ELEM( *m, float, M_SX, S_SX ) = 1;
    CV_MAT_ELEM( *m, float, M_SY, S_SY ) = 1;
    cvSetIdentity( kalman->transition_matrix );
    setProcessNoiseCov( sign );
    /* Measures */
    for( int j = 0; j < FRAMES_MAX; j++)
        sign->measures[j].vector = cvCreateMat( NUMBER_OF_MEASURES, 1, CV_32FC1 );

    resetSign( sign );
  }

  /* Allocate memory for detections. */
  /* For Haar sequence */
  tracking->haarDetectionStorage = cvCreateMemStorage(0);
  /* For reformatted sequence */
  tracking->detectionStorage = cvCreateMemStorage(0);
  tracking->detectionNoiseCov = cvCreateMat( NUMBER_OF_MEASURES,
                                             NUMBER_OF_MEASURES,
                                             CV_32FC1 );
  /* Load cascade of Haar classifiers. */
  tracking->cascade = cvLoadHaarClassifierCascade( HAAR_CASCADE_PATH,
                                                   HAAR_IMAGE_SIZE );
  //tracking->cascade = (CvHaarClassifierCascade*) cvLoad( HAAR_CASCADE_PATH );
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

void resetSign( Sign* sign )
{
  FUNCTION_BEGIN;
  /* Measures and patches */
  sign->age = 0;
  sign->numberOfMissingMeasures = 0;
  sign->numberOfPatches = 0;
  sign->numberOfPredictions = 0;
  sign->numberOfCorrections = 0;  
  for( int j = 0; j < FRAMES_MAX; j++)
  {
    Measure* measure = &sign->measures[j];
    measure->exist = 0;
    measure->confidence = 0;
    Patch* patch = &sign->patches[j];
    patch->exist = 0;
    patch->confidence = 0;
  }
  /* Kalman */
  CvKalman* kalman = sign->kalman;
  cvSetIdentity( kalman->transition_matrix );
  cvSetZero( kalman->error_cov_post );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

int freeTracking( Tracking* tracking)
{
  FUNCTION_BEGIN;
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];

    /* Kalman */
    if( sign->kalman != NULL )
    {
      PRINT9("free kalman structure of sign %d", i);
      cvReleaseKalman( &sign->kalman ); 
    }

    /* Measures and patches */
    for( int j = 0; j < FRAMES_MAX; j++)
        cvReleaseMat( &sign->measures[j].vector );
  }

  /* Detection */
  cvReleaseMemStorage( &tracking->haarDetectionStorage );
  cvReleaseMemStorage( &tracking->detectionStorage );
  cvReleaseMat( &tracking->detectionNoiseCov );

  /* Cascade of Haar classifiers. */
  cvReleaseHaarClassifierCascade( &tracking->cascade );
  
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

void setTransitionMatrix( Sign* sign, float dt )
{
  FUNCTION_BEGIN;
  /* 8-state dynamic model, 8x8 transition matrix. Constant
   * acceleration model for position. */
  CvMat* a = sign->kalman->transition_matrix;
  //cvSetIdentity( a ); // already made at the time of tracking initialization
  float c = dt * dt / 2;
  CV_MAT_ELEM( *a, float, S_X, S_VX ) = dt; // CV_MAT_ELEM faster than cvmSet? I think yes.
  CV_MAT_ELEM( *a, float, S_X, S_AX ) = c;
  CV_MAT_ELEM( *a, float, S_VX, S_AX ) = dt;
  CV_MAT_ELEM( *a, float, S_Y, S_VY ) = dt;
  CV_MAT_ELEM( *a, float, S_Y, S_AY ) = c;
  CV_MAT_ELEM( *a, float, S_VY, S_AY ) = dt;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void setProcessNoiseCov( Sign* sign )
{
  FUNCTION_BEGIN;
  /* 8-state dynamic model, 8x8 covariance matrix. */
  CvMat* q = sign->kalman->process_noise_cov;
  cvSetZero( q );
  CV_MAT_ELEM( *q, float, S_AX, S_AX ) = 1e4;
  CV_MAT_ELEM( *q, float, S_AY, S_AY ) = 1e4;
  CV_MAT_ELEM( *q, float, S_SX, S_SX ) = 1e2;
  CV_MAT_ELEM( *q, float, S_SY, S_SY ) = 1e2;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void setMeasurementNoiseCov( Tracking* tracking, Sign* sign, double scale )
{
  FUNCTION_BEGIN;
  /* 4-vector measurement, 4x4 covariance matrix */
  CvMat* m = sign->kalman->measurement_noise_cov;
  cvCopy( tracking->detectionNoiseCov, m );
  if( scale != 1)
      cvConvertScale( m, m, scale );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void predictState( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  double dt = video->dts[video->ti];
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];
    if( sign->age > 0 ) // only tracked signs
    {
      /* New sign require Kalman filter initialization. */
      if( sign->age == 1 )
          initKalman( sign, video );
      setTransitionMatrix( sign, dt );
      cvKalmanPredict( sign->kalman, 0 );
      sign->numberOfPredictions++;
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawPredictions( Tracking* tracking, Video* video, IplImage* picture,
                      CvScalar color, int pointRadius, double ellipseScale )
{
  FUNCTION_BEGIN;
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    if( sign->numberOfPredictions > 0 )
    {
      CvKalman* kalman = sign->kalman;
      CvMat* m = kalman->state_pre;
      CvPoint2D32f position_pre = cvPoint2D32f( ((float*)(m->data.ptr + m->step*S_X))[0],
                                                ((float*)(m->data.ptr + m->step*S_Y))[0] );
      CvScalar ellipse;
      covEllipse( kalman->error_cov_pre, 0, 1, &ellipse);
      cvCircle( picture, cvPointFrom32f( position_pre ), pointRadius, color, -1 );
      CvSize axes = cvSize( (int) (ellipseScale * ellipse.val[0]),
                            (int) (ellipseScale * ellipse.val[1]) );
      drawEllipse( picture, cvPointFrom32f( position_pre ), axes,
                   -ellipse.val[2]*180/M_PI, 0, 360, color );
    }
  }
  FUNCTION_END;
}   

/* -------------------------------------------------------------------------- */

void detectPoints( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  setDetectionNoiseCov( tracking );
  /* Look for all maxima (we are tracking points on a black
   * background) in the current frame. Results are put in a
   * sequence. */
  int fi = video->fi;
  IplImage* frame = video->grayscaleFrames[fi];
  cvThreshold( frame, frame, 10, 255, CV_THRESH_BINARY );
  IplImage* mask = video->mask;
  cvSet( mask, cvScalarAll(255) );
  //cvSetImageCOI( frame, 1 ); // for 3-channel frame
  double minVal, maxVal;
  
  CvPoint minLoc, maxLoc;
  tracking->detections = cvCreateSeq( CV_SEQ_ELTYPE_POINT, sizeof(CvSeq),
                                      sizeof(CvPoint), tracking->detectionStorage );
  CvSeq* detections = tracking->detections;
  
  while(1)
  {
    /* Locate maximum and push it into the sequence */
    cvMinMaxLoc( frame, &minVal, &maxVal, &minLoc, &maxLoc, mask );
    if( maxVal == 0 )
        break;
    PRINT9("maximum value = %.1f, location = (%d,%d)", maxVal, maxLoc.x, maxLoc.y );
    cvSeqPush( detections, &maxLoc );
    /* Mask current maximum location for next measures */
    #define RECT_W 21 // should be odd
    #define RECT_H 21 // should be odd
    CvRect signPatch = cvRect( maxLoc.x - (RECT_W-1)/2, maxLoc.y - (RECT_H-1)/2,
                               RECT_W, RECT_H);
    #undef RECT_W
    #undef RECT_H
    cvSetImageROI( mask, signPatch );
    cvSetZero( mask );
    cvResetImageROI( mask );
    /* Draw position and patch to verify */
/*     IplImage* display = video->display; */
/*     cvCircle( display, maxLoc, 2, CV_WHITE, -1 ); */
/*     cvRectangle( display, cvPoint(signPatch.x, signPatch.y), */
/*                  cvPoint(signPatch.x + signPatch.width, signPatch.y + signPatch.height), */
/*                  CV_WHITE ); */
  }
  //cvSetImageCOI( frame, 0 ); // for 3-channel frame
  if( detections->total > SIGNS_MAX )
  {
    PRINT_ERR( TERM_RED "careful! Number of detections (%d) is greater "
               "than number of maximum tracked signs (%d)." TERM_RESET,
               detections->total, SIGNS_MAX);
  }
  else
  {
    PRINT9( "there are %d detection(s) in current frame", detections->total );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void detectSigns( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  setDetectionNoiseCov( tracking );
  int fi = video->fi;
  IplImage* frame = video->grayscaleFrames[fi];
  CvSeq* haarDetections = cvHaarDetectObjects( frame, tracking->cascade,
                                               tracking->haarDetectionStorage,
                                               HAAR_SCALE,HAAR_NEIGHBORS,
                                               CV_HAAR_DO_CANNY_PRUNING,
                                               HAAR_IMAGE_SIZE_MINIMUM_DETECTION );
  if( haarDetections->total > SIGNS_MAX )
  {
    PRINT_ERR( TERM_RED "careful! Number of detections (%d) is greater "
               "than number of maximum tracked signs (%d)." TERM_RESET,
               haarDetections->total, SIGNS_MAX);
  }
  else
  {
    PRINT9( "there are %d detection(s) in current frame", haarDetections->total );
  }

  /* Reformat the sequence. */
  CvSeq* detections = cvCreateSeq( 0, sizeof(CvSeq), sizeof(Detection),
                                   tracking->detectionStorage );
  for( int i = 0; i < haarDetections->total; i++ )
  {
    CvAvgComp* haarDetection = (CvAvgComp*) cvGetSeqElem( haarDetections, i );
    CvRect* rect = &haarDetection->rect;
    Detection detection;
    detection.rect = haarDetection->rect;
    detection.neighbors = haarDetection->neighbors;
    detection.topLeftPoint = cvPoint( rect->x, rect->y );
    int xOffset = rect->width;
    int yOffset = rect->height;
    detection.bottomRightPoint = cvPoint( rect->x + xOffset,
                                          rect->y + yOffset );
    detection.center = cvPoint2D32f( rect->x + xOffset / 2.,
                                     rect->y + yOffset / 2. );
    cvSeqPush( detections, &detection );
  }
  tracking->detections = detections;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void keepGoodDetections( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  int fi = video->fi;
  IplImage* frame = video->frames[fi];
  CvSeq* detections = tracking->detections;
  for( int i = 0; i < detections->total; i++ )
  {
    Detection* detection = (Detection*) cvGetSeqElem( detections, i );
    if( isRedSign(frame, detection->rect, video) == 0
        || isGoodSize(detection->rect) == 0)
    {
      cvSeqRemove( detections, i );
      i--;
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void setDetectionNoiseCov( Tracking* tracking )
{
  FUNCTION_BEGIN;
  /* At present, it's always the same and for all signs. */
  CvMat* m = tracking->detectionNoiseCov;
  cvSetZero(m);
  CV_MAT_ELEM( *m, float, M_X, M_X ) = 4;
  CV_MAT_ELEM( *m, float, M_Y, M_Y ) = 9;
  CV_MAT_ELEM( *m, float, M_SX, M_SX ) = 16;
  CV_MAT_ELEM( *m, float, M_SY, M_SY ) = 16;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawDetectedPoints( Tracking* tracking, Video* video, IplImage* picture,
                         CvScalar color, int pointRadius, double ellipseScale )
{
  FUNCTION_BEGIN;
  CvSeq* detections = tracking->detections;
  for( int i = 0; i < detections->total; i++ )
  {
    CvPoint position = *(CvPoint*) cvGetSeqElem( detections, i );
    cvCircle( video->display, position, pointRadius, color, -1 );

    /* Error ellipse */
    CvScalar ellipse;
    covEllipse( tracking->detectionNoiseCov, 0, 1, &ellipse);
    CvSize axes = cvSize( (int) (ellipseScale * ellipse.val[0]),
                          (int) (ellipseScale * ellipse.val[1]) );
    drawEllipse( picture, position, axes,
                 -ellipse.val[2]*180/M_PI, 0, 360, color );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawDetectedSigns( Tracking* tracking, Video* video, IplImage* picture,
                        CvScalar color, int thickness, int pointRadius,
                        double ellipseScale )
{
  FUNCTION_BEGIN;
  CvSeq* detections = tracking->detections;
  for( int i = 0; i < detections->total; i++ )
  {
    Detection* detection = (Detection*) cvGetSeqElem( detections, i );
    /* Draw rectangle */
    cvRectangle( picture, detection->topLeftPoint,
                 detection->bottomRightPoint, color, thickness );
    /* Draw center */
    cvCircle( picture, cvPointFrom32f(detection->center), pointRadius, color, -1 );
    /* Error ellipse for center */
    CvScalar ellipse;
    covEllipse( tracking->detectionNoiseCov, 0, 1, &ellipse);
    CvSize axes = cvSize( (int) (ellipseScale * ellipse.val[0]),
                          (int) (ellipseScale * ellipse.val[1]) );
    drawEllipse( picture, cvPointFrom32f(detection->center), axes,
                 -ellipse.val[2]*180/M_PI, 0, 360, color );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawMeasures( Tracking* tracking, Video* video, IplImage* picture,
                   CvScalar color1, int thickness1, int pointRadius1,
                   CvScalar color2, int thickness2, int pointRadius2,
                   double ellipseScale )
{
  FUNCTION_BEGIN;
  CvScalar color;
  int thickness, pointRadius;
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    Measure *measure = &sign->measures[video->fi];

/*     printf("\nPatches existence of sign %d: ", i); */
/*     for( int j = 0; j < FRAMES_MAX; j++ ) */
/*         printf("%d ", sign->patches[j].exist); */
    
    if( measure->exist == 1 )
    {
      if( sign->patches[video->fi].exist != 0 )
      {
        color = color2;
        thickness = thickness2;
        pointRadius = pointRadius2;
      }
      else
      {
        color = color1;
        thickness = thickness1;
        pointRadius = pointRadius1;
      }
      CvMat* vec = measure->vector;
      /* Center */
      CvPoint2D32f position;
      position.x = cvmGet( vec, M_X, 0 );
      position.y = cvmGet( vec, M_Y, 0 );
      cvCircle( picture, cvPointFrom32f(position), pointRadius, color, -1 );
      /* Rectangle */
      CvPoint topLeftPoint, bottomRightPoint;
      topLeftPoint.x = (int) round( position.x - cvmGet( vec, M_SX, 0 ) / 2 );
      topLeftPoint.y = (int) round( position.y - cvmGet( vec, M_SY, 0 ) / 2 );
      bottomRightPoint.x = (int) round( position.x + cvmGet( vec, M_SX, 0 ) / 2 );
      bottomRightPoint.y = (int) round( position.y + cvmGet( vec, M_SY, 0 ) / 2 );
      cvRectangle( picture, topLeftPoint, bottomRightPoint, color, thickness );
      /* Error ellipse for center */
      CvScalar ellipse;
      covEllipse( sign->kalman->measurement_noise_cov, 0, 1, &ellipse);
      CvSize axes = cvSize( (int) (ellipseScale * ellipse.val[0]),
                            (int) (ellipseScale * ellipse.val[1]) );
      drawEllipse( picture, cvPointFrom32f(position), axes,
                   -ellipse.val[2]*180/M_PI, 0, 360, color );
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawLabels( Tracking* tracking, Video* video, IplImage* picture )
{
  FUNCTION_BEGIN;
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    Measure* measure = &sign->measures[video->fi];
    if( sign->age > 0 )
    {
      /* Get position. */
      CvPoint position;
      CvMat* vec = measure->vector;
      position.x = (int) cvmGet( vec, M_X, 0 );
      position.y = (int) cvmGet( vec, M_Y, 0 );

      /* Put label. */
      char label[16];
      sprintf( label, "%d:%d,%d", i, sign->age, sign->numberOfMissingMeasures );
      putText( picture, label, cvPoint(position.x, position.y - 30),
               &video->fonts.plainSmall, CV_BLACK, CV_WHITE );
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void dataAssociation( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  /* Association vector. associations[i]=j means that i-th signs is
   * associated with the j-th detection in the sequence. j=-1 means no
   * association. */
  int associations[SIGNS_MAX];
  for( int i = 0; i < SIGNS_MAX; i++ ) // initialization
      associations[i] = -1;

  CvSeq* detections = tracking->detections;
  
  /* Compute distances between tracked signs and detections. */
  CvMat* distanceMatrix;
  computeDistances( tracking, &distanceMatrix );
  if( distanceMatrix == NULL )
  {
    PRINT9("skip distances computing");
    goto after_distances_computing;
  }
  
  /* Then we try to associate each prediction with a detection. */
  PRINT9("try to associate each prediction with a detection");
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    /* Associate the prediction with the nearest detection if distance
     * is less than a threshold (it's the gating). */
    if( sign->age > 0 ) // is it a tracked sign ?
    {
      int j = 0;
      associations[i] = j;
      double minDist = cvmGet( distanceMatrix, i, j );
      for( j = 1; j < detections->total; j++ )
      {
        double dist = cvmGet( distanceMatrix, i, j );
        if( dist < minDist )
        {
          minDist = dist;
          associations[i] = j;
        }
      }
      /* Gating */
      int asso = associations[i];
      if( asso != -1 && cvmGet(distanceMatrix, i, asso) > DISTANCE_MAX )
          associations[i] = -1;
    }
  }
#if VERBOSE >= 9
  printArray( (void*)associations, 1, SIGNS_MAX, CV_32SC1,
              "association vector" );
#endif
  
  /* If a detection is associated with more than one prediction, we
   * keep the pair with the smallest distance. */
  PRINT9("keep good pairs");
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign1 = &tracking->signs[i];
    if( sign1->age > 0 ) // is it a tracked sign ?
    {
      int index1 = associations[i];
      for( int j = 0; j < SIGNS_MAX; j++ )
      {
        Sign* sign2 = &tracking->signs[j];
        if( sign2->age > 0 ) // is it a tracked sign ?
        {
          int index2 = associations[j];
          if( index2 != -1 && index2 == index1 )
          {
            double dist1 = cvmGet( distanceMatrix, i, index1 );
            double dist2 = cvmGet( distanceMatrix, j, index1 );
            if( dist2 < dist1 )
            {
              /* No detection associated with prediction of sign1. */
              associations[i] = -1;
            }
          }
        }
      }
    }
  }
  cvReleaseMat( &distanceMatrix );

  
  after_distances_computing:
#if VERBOSE >= 9
  printArray( (void*)associations, 1, SIGNS_MAX, CV_32SC1,
              "association vector before creating new signs" );
#endif
  
  /* We need a lookup table telling if a detection is associated or
   * no. */
  int associatedDetections[DETECTIONS_MAX];
  /* Initialize the LUT. */
  for( int i = 0; i < DETECTIONS_MAX; i++ )
      associatedDetections[i] = -1;
  /* Create the LUT. */
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    int asso = associations[i];
    if( asso != -1 )
        associatedDetections[asso] = asso;
  }
#if VERBOSE >= 9
  if( detections->total )
      printArray( (void*)associatedDetections, 1, detections->total, CV_32SC1,
                  "associated detections flags" );
#endif

  /* Create new signs with unassociated detections in the
   * sequence. */
  PRINT9("Create new signs with unassociated detections in the sequence");
  int freeSignIndex = 0;
  for( int i = 0; i < detections->total; i++ )
  {
    if( associatedDetections[i] == -1 )
    {
      /* Look for a free sign (not tracked yet). */
      int j;
      for( j = freeSignIndex; tracking->signs[j].age != 0 && j < SIGNS_MAX; j++ )
      if( j == SIGNS_MAX )
          break;
      associations[j] = i;
      freeSignIndex = j + 1;
    }
  }

#if VERBOSE >= 9
  printArray( (void*)associations, 1, SIGNS_MAX, CV_32SC1,
              "association vector after creating new signs" );
#endif
  
  /* Now, association is finished. We have to replace indexes by real
   * measures and stop tracking signs whose number of missing measures
   * is too high. */
  PRINT9("replace association indexes by real measures");
  convertIndexToMeasure( tracking, video, associations );

  
  /* Haar sequence (from cvhaar.cpp) */
/*   CvMemStorage* storage = cvCreateMemStorage(0); */
/*   CvSeq* signsFromHaar = cvCreateSeq(0, sizeof(CvSeq), */
/*                                       sizeof(CvAvgComp), storage ); */
/*   signsFromHaar = cvHaarDetectObjects( frame, cascade, storage, HAAR_SCALE, */
/*                                        HAAR_NEIGHBORS, CV_HAAR_DO_CANNY_PRUNING ); */
/*   for( int i = 0; i < signsFromHaar->total; i++ ) */
/*   { */
/*     /\* Extract the rectangles only *\/ */
/*     CvRect signPatch = *(CvRect*) cvGetSeqElem( signsFromHaar, i ); */
/*     cvRectangle( display, cvPoint(signPatch.x, signPatch.y), */
/*                  cvPoint(signPatch.x + signPatch.width, signPatch.y + signPatch.height), */
/*                  CV_GREEN ); */
/*   } */
/*   cvReleaseMemStorage( &storage ); */
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void computeDistances( Tracking* tracking, CvMat** distanceMatrix )
{
  FUNCTION_BEGIN;
  CvSeq* detections = tracking->detections;
  if( detections->total == 0 || tracking->numberOfSigns == 0 )
  {
    PRINT9("no detection or no tracked sign, so no distance to compute.");
    *distanceMatrix = NULL;
    FUNCTION_END;
    return;
  }
  
  *distanceMatrix = cvCreateMat( SIGNS_MAX, detections->total, CV_64FC1 );
  cvSet( *distanceMatrix, cvScalarAll(-1) ); // for debugging (not necessary)

#ifdef NEAREST_NEIGHBOR_SQUARE_EUCLIDIAN_DISTANCE
  /* Distance = x^2 + y^2 */
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    if( sign->age > 0 ) // is it a tracked sign ?
    {
      CvMat* m = sign->kalman->state_pre;
      CvPoint2D32f predictedPosition = cvPoint2D32f( ((float*)(m->data.ptr + m->step*S_X))[0],
                                                     ((float*)(m->data.ptr + m->step*S_Y))[0] );
      for( int j = 0; j < detections->total; j++ )
      {
        Detection* detection = (Detection*) cvGetSeqElem( detections, j );
        double dist = icvSqDist2D32f( predictedPosition, detection->center );
        CV_MAT_ELEM( **distanceMatrix, double, i, j ) = dist;
      }
    }
  }
#else
#ifdef NEAREST_NEIGHBOR_MAHALANOBIS_DISTANCE
  /* Distance = X^T * COV^-1 * X
   * where X = detection - prediction
   * and COV = M * COV_PREDICTION * M^T + COV_MEASURE.
   * View Computer Vision (Forsyth and Ponce) p.389. */
  CvMat* tmpMatrix = cvCreateMat( NUMBER_OF_STATES, NUMBER_OF_STATES, CV_32FC1 );
  CvMat* covMatrix = cvCreateMat( NUMBER_OF_MEASURES, NUMBER_OF_MEASURES, CV_32FC1 );
  CvMat* inverseCovMatrix = cvCreateMat( NUMBER_OF_MEASURES, NUMBER_OF_MEASURES, CV_32FC1 );
  CvMat* predictionVec = cvCreateMat( NUMBER_OF_MEASURES, 1, CV_32FC1 );
  CvMat* detectionVec = cvCreateMat( NUMBER_OF_MEASURES, 1, CV_32FC1 );
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    if( sign->age > 0 ) // is it a tracked sign ?
    {
      CvKalman* kalman = sign->kalman;
      /* Prediction vector. */
      cvGEMM( kalman->measurement_matrix, kalman->state_pre, 1,
              0, 1, predictionVec, 0 );
      //printMatrix( predictionVec, "predictionVec" );
      /* First term (common part) of covariance matrix. */
      doubleProduct32f( kalman->measurement_matrix, kalman->error_cov_pre,
                        kalman->measurement_matrix, covMatrix );
      //printMatrix( covMatrix, "covMatrix" );
      
      for( int j = 0; j < detections->total; j++ )
      {
        /* Detection vector. */
        Detection* detection = (Detection*) cvGetSeqElem( detections, j );
        CV_MAT_ELEM( *detectionVec, float, M_X, 0 ) = detection->center.x;
        CV_MAT_ELEM( *detectionVec, float, M_Y, 0 ) = detection->center.y;
        CV_MAT_ELEM( *detectionVec, float, M_SX, 0 ) = detection->rect.width;
        CV_MAT_ELEM( *detectionVec, float, M_SY, 0 ) = detection->rect.height;
        //printMatrix( detectionVec, "detectionVec" );
        /* Add second part of covariance matrix and invert it. */
        cvAdd( covMatrix, kalman->measurement_noise_cov, covMatrix );
        //printMatrix( covMatrix, "covMatrix" );
        cvInvert( covMatrix, inverseCovMatrix, CV_SVD );
        //printMatrix( inverseCovMatrix, "inverseCovMatrix" );
        /* Compute distance. */
        double dist = cvMahalanobis( predictionVec, detectionVec, inverseCovMatrix );
        CV_MAT_ELEM( **distanceMatrix, double, i, j ) = dist;
      }
    }
  }
  cvReleaseMat( &tmpMatrix );
  cvReleaseMat( &covMatrix );
  cvReleaseMat( &inverseCovMatrix );
  cvReleaseMat( &predictionVec );
  cvReleaseMat( &detectionVec );
#else
  PRINT_ERR("No DATA_ASSOCIATION_DISTANCE defined");
  exit(-1);
#endif
#endif
  
#if VERBOSE >= 9
  printMatrix( *distanceMatrix, "distanceMatrix" );
#endif
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void convertIndexToMeasure( Tracking* tracking, Video* video, int* associations )
{
  FUNCTION_BEGIN;
  int fi = video->fi;
  CvSeq* detections = tracking->detections;
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    Measure* measure = &sign->measures[fi];
    int detectionIndex = associations[i];
    if( detectionIndex != -1 ) // association exist
    {
      /* Save the association as a measure for the current sign. */
      sign->age++;
      sign->numberOfMissingMeasures = 0;
      CvMat* z = measure->vector;
      measure->exist = 1;
      Detection* detection = (Detection*) cvGetSeqElem( detections, detectionIndex );

      /* Put detection into sign's measure and set Kalman measurement
       * noise covariance.*/
      CV_MAT_ELEM( *z, float, M_X, 0 ) = detection->center.x;
      CV_MAT_ELEM( *z, float, M_Y, 0 ) = detection->center.y;
      CV_MAT_ELEM( *z, float, M_SX, 0 ) = detection->rect.width;
      CV_MAT_ELEM( *z, float, M_SY, 0 ) = detection->rect.height;
      setMeasurementNoiseCov( tracking, sign );
    
      if( sign->age == 1 ) // new tracked sign
      {
        PRINT9("sign %d is a new tracked sign", i);
        tracking->numberOfSigns++;
      }
      else
      {
        PRINT9("continue tracking sign %d (with an association)", i);
      }
    }
    else // association doesn't exist
    {
      if( sign->age > 0 ) // is it a tracked sign ?
      {
        /* Tracked sign has no measure in this frame. We have to
         * remember that because the dynamic model have to predict the
         * future state without any measure. If the sign has no
         * measure for several successive frames, we stop tracking
         * it. */
        measure->exist = 0;
        sign->numberOfMissingMeasures++;
        if( sign->numberOfMissingMeasures > MISSING_MEASURES_MAX )
        {
          PRINT9("stop tracking sign %d", i);
          resetSign( sign );
          tracking->numberOfSigns--;
        }
        else
        {
          PRINT9("continue tracking sign %d (without any association)", i);
          sign->age++;
        }
      }
      else
      {
        /* Nothing to do. It's a no tracked sign which has no
         * association. */
        PRINT9("Nothing to do. It's a no tracked sign which has no association");
      }
    }
  }
/* #if VERBOSE >= 9 */
/*   printTracking( tracking, NULL, fi, 1 ); */
/* #endif */
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void correctState( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];
    if( sign->numberOfPredictions > 0 )
    {
      CvKalman* kalman = sign->kalman;
      Measure* measure = &sign->measures[video->fi];
      if( sign->measures[video->fi].exist == 0 )
      {
        /* redondant debugging */
        if( sign->numberOfMissingMeasures == 0 )
        {          
          PRINT_ERR("Something wrong! You should never see this text!!");
          exit(-1);
        }
        /* No measure in this frame, so we put all confidence in the
         * dynamic model by setting measure equal to the prediction
         * but we increase measurement noise covariance because this
         * is not a real measure, thus correction error covariance
         * will increase and then also next prediction error
         * covariance. */
        cvMatMulAdd( kalman->measurement_matrix, kalman->state_pre, 0, measure->vector );
        setMeasurementNoiseCov( tracking, sign, 4*sign->numberOfMissingMeasures );
      }
      cvKalmanCorrect( kalman, measure->vector );
      sign->numberOfCorrections++;
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawCorrections( Tracking* tracking, Video* video, IplImage* picture,
                      CvScalar color, int pointRadius, double ellipseScale )
{
  FUNCTION_BEGIN;
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];
      
    if( sign->numberOfCorrections > 0 )
    {
      CvKalman* kalman = sign->kalman;
      CvMat* m = kalman->state_post;
      CvPoint2D32f position_post = cvPoint2D32f( ((float*)(m->data.ptr + m->step*S_X))[0],
                                                 ((float*)(m->data.ptr + m->step*S_Y))[0] );
      cvCircle( picture, cvPointFrom32f( position_post ), pointRadius, color, -1 );
      CvScalar ellipse;
      covEllipse( kalman->error_cov_post, 0, 1, &ellipse);
      CvSize axes = cvSize( (int) (ellipseScale * ellipse.val[0]),
                            (int) (ellipseScale * ellipse.val[1]) );
      drawEllipse( picture, cvPointFrom32f( position_post ), axes,
                   -ellipse.val[2]*180/M_PI, 0, 360, color );
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void initKalman( Sign* sign, Video* video )
{
  FUNCTION_BEGIN;
  /* Review: state prediction: x(-,k) = A(k) * x(+,k-1). */
  /* Corrected state equal measurement.Prediction must be at the same
   * position than measurement because we don't know the direction in
   * which the sign is moving. (could be improved using optical flow
   * to estimate velocity). So we have to clear velocity and
   * acceleration in x(+,k-1). */
  
  CvKalman* kalman = sign->kalman;
  CvMat* statePost = kalman->state_post;
  /* We have to use the measure of *previous* frame because initKalman
   * is called after a new frame acquisition (because we need the new
   * time index to set error_cov_post below). */
  int fi = video->fi;
  CvMat* z = sign->measures[fi ? fi-1 : FRAMES_MAX-1].vector;
  cvSetZero( statePost );
  CV_MAT_ELEM( *statePost, float, S_X, 0 ) = cvmGet( z, M_X, 0 );
  CV_MAT_ELEM( *statePost, float, S_Y, 0 ) = cvmGet( z, M_Y, 0 );
  CV_MAT_ELEM( *statePost, float, S_SX, 0 ) = cvmGet( z, M_SX, 0 );
  CV_MAT_ELEM( *statePost, float, S_SY, 0 ) = cvmGet( z, M_SY, 0 );
  /* Review: prediction error cov: P(-,k) = A(k) * P(+,k-1) * A(k)^T +
   * Q = P(+,k-1) + Q, because A = 1. */
  /* Longer time between frames is, most important must be the
   * prediction error covariance. */
  CvMat* errorCovPost = kalman->error_cov_post;
  cvSetZero( errorCovPost );
  double dt = video->dts[video->ti];
  CV_MAT_ELEM( *errorCovPost, float, S_X, S_X ) = dt * 1e2;
  CV_MAT_ELEM( *errorCovPost, float, S_Y, S_Y ) = dt * 1e2;
  CV_MAT_ELEM( *errorCovPost, float, S_SX, S_SX ) = dt * 1e2;
  CV_MAT_ELEM( *errorCovPost, float, S_SY, S_SY ) = dt * 1e2;
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

int isRedSign( const IplImage* src, CvRect roi, Video* video )
{
  FUNCTION_BEGIN;

  if( src->nChannels != 3 )
  {
    PRINT_ERR( "source image has %d channel(s) but 3 is required", src->nChannels );
    FUNCTION_END;
    return -1;
  }
  
  //CvRect roi = cvGetImageROI( src );
  int xOffset = roi.x;
  int yOffset = roi.y;
  int width = roi.width;
  int height = roi.height;

  /* We cut image into nine rectangles (3x3). We test rectangles (2,0)
   * and (2,2) by counting the proportion of red pixels. */
  int hTicks[4] = {xOffset, xOffset + width / 3, xOffset + 2 * width / 3, xOffset + width};
  int vTicks[4] = {yOffset, yOffset + height / 3, yOffset + 2 * height / 3, yOffset + height};
  double area = width * height / 9.;
  double numberOfRed = 0;
  IplImage* picture = video->display[0];
  
  const double proportion = 0.25; /* minimum proportion or red in cells */
  
  /* Cell 0,1 */
  for( int i = vTicks[0]; i < vTicks[1]; i++ )
  {
    for( int j = hTicks[1]; j < hTicks[2]; j++ )
    {
      uchar* srcPixel = &((uchar*)(src->imageData + src->widthStep*i))[j*3];
      numberOfRed += isRed( srcPixel );
      /* Draw */
      uchar* pixel = &((uchar*)(picture->imageData + picture->widthStep*i))[j*3];
      pixel[0] /= 2;
      pixel[1] /= 2;
    }
  }
  if( numberOfRed / area > proportion )
  {
    FUNCTION_END;
    return 1;
  }
  
  /* Cell 2,0 */
  numberOfRed = 0;
  for( int i = vTicks[2]; i < vTicks[3]; i++ )
  {
    for( int j = hTicks[0]; j < hTicks[1]; j++ )
    {
      uchar* srcPixel = &((uchar*)(src->imageData + src->widthStep*i))[j*3];
      numberOfRed += isRed( srcPixel );
      /* Draw */
      uchar* pixel = &((uchar*)(picture->imageData + picture->widthStep*i))[j*3];
      pixel[0] /= 2;
      pixel[1] /= 2;
    }
  }
  if( numberOfRed / area > proportion )
  {
    FUNCTION_END;
    return 1;
  }
  
  /* Cell 2,2 */
  numberOfRed = 0;
  for( int i = vTicks[2]; i < vTicks[3]; i++ )
  {
    for( int j = hTicks[2]; j < hTicks[3]; j++ )
    {
      uchar* srcPixel = &((uchar*)(src->imageData + src->widthStep*i))[j*3];
      numberOfRed += isRed( srcPixel );
      /* Draw */
      uchar* pixel = &((uchar*)(picture->imageData + picture->widthStep*i))[j*3];
      pixel[0] /= 2;
      pixel[1] /= 2;
    }
  }
  if( numberOfRed / area > proportion )
  {
    FUNCTION_END;
    return 1;
  }

  FUNCTION_END;
    return 0;
}

/* -------------------------------------------------------------------------- */

int isGoodSize( CvRect roi )
{
  FUNCTION_BEGIN;
  const int threshold = 100;
  if( roi.width < threshold && roi.height < threshold )
  {
    FUNCTION_END;
      return 1;
  }
  else
  {
    FUNCTION_END;
    return 0;
  }
}

/* -------------------------------------------------------------------------- */

void savePatches( Tracking* tracking, Video* video )
{
  FUNCTION_BEGIN;
  if( video->numberOfGrabs < AGE_TO_BE_A_SIGN )
  {
    FUNCTION_END;
    return;
  }
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];
    int precedentPatchExist = sign->patches[INDEX_CIRC_BUFFER(video->fi-1, FRAMES_MAX)].exist;
    if( sign->age >= AGE_TO_BE_A_SIGN && sign->measures[video->fi].exist != 0
        || precedentPatchExist != 0 )
    {
      /* If precedent patch doesn't exist, we have to save all
       * preceding measures, else we have to save only the current
       * measure. */
      /* TO DO: save all precedent images og the tracks, not only the
       * last AGE_TO_BE_A_SIGN (thera are most images in certains
       * cases). */
      int jMin;
      if( precedentPatchExist == 0 )
          jMin = 0;
      else
          jMin = AGE_TO_BE_A_SIGN - 1;
      for( int j = AGE_TO_BE_A_SIGN - 1; j >= jMin; j-- )
      {
        int patchIndex = INDEX_CIRC_BUFFER(video->fi-(AGE_TO_BE_A_SIGN-1)+j, FRAMES_MAX);
        Patch* patch = &sign->patches[patchIndex];
        patch->exist = 1;
        sign->numberOfPatches++;
        /* Save patch. */
        CvPoint* topLeftPoint = &patch->topLeftPoint;
        CvPoint* bottomRightPoint = &patch->bottomRightPoint;
        CvMat* vec = sign->measures[patchIndex].vector;
        CvPoint2D32f position;
        position.x = cvmGet( vec, M_X, 0 );
        position.y = cvmGet( vec, M_Y, 0 );
        topLeftPoint->x = (int) round( position.x - cvmGet( vec, M_SX, 0 ) / 2 );
        topLeftPoint->y = (int) round( position.y - cvmGet( vec, M_SY, 0 ) / 2 );
        bottomRightPoint->x = (int) round( position.x + cvmGet( vec, M_SX, 0 ) / 2 );
        bottomRightPoint->y = (int) round( position.y + cvmGet( vec, M_SY, 0 ) / 2 );
      }  
    }
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void drawPatches( Tracking* tracking, Video* video, CvScalar color,
                  int thickness )
{
  FUNCTION_BEGIN;
  if( video->numberOfGrabs < AGE_TO_BE_A_SIGN )
  {
    FUNCTION_END;
    return;
  }
  for( int i = 0; i < SIGNS_MAX; i++)
  {
    Sign* sign = &tracking->signs[i];
    if( sign->age >= AGE_TO_BE_A_SIGN )
    {
      int jMin;
      if( sign->age == AGE_TO_BE_A_SIGN )
          jMin = 0;
      else
          jMin = AGE_TO_BE_A_SIGN - 1;
      for( int j = AGE_TO_BE_A_SIGN - 1; j >= jMin; j-- )
      {
        int patchIndex = INDEX_CIRC_BUFFER(video->fi-(AGE_TO_BE_A_SIGN-1)+j, FRAMES_MAX);
        Patch* patch = &sign->patches[patchIndex];
        if( patch->exist != 0 )
        {
          IplImage* picture = video->frames[patchIndex];
          /* Draw patch. */
          cvRectangle( picture, patch->topLeftPoint, patch->bottomRightPoint,
                       color, thickness );
          /* Put label. */
          char label[16];
          sprintf( label, "%d", i );
          putText( picture, label, patch->topLeftPoint,
                   &video->fonts.plainSmall, CV_BLACK, CV_WHITE );
        }
      }  
    }
  }
  FUNCTION_END;
}
