#ifndef _TRACKING_H
#define _TRACKING_H

#include <cv.h>
#include "video.h"
#include "color.h"

/* Kalman */
#define NUMBER_OF_STATES 8
#define NUMBER_OF_MEASURES 4
#define NUMBER_OF_CONTROLS 0

/* State index into matrices. */
#define S_X 0 // horizontal position
#define S_Y 1 // vertical position
#define S_VX 2 // horizontal velocity
#define S_VY 3 // vertical velocity
#define S_AX 4 // horizontal acceleration
#define S_AY 5 // vertical acceleration
#define S_SX 6 // horizontal size
#define S_SY 7 // vertical size

/* Measure index into matrices. */
#define M_X 0 // horizontal position
#define M_Y 1 // vertical position
#define M_SX 2 // horizontal size
#define M_SY 3 // vertical size

/* Maximum detected objects per frame. */
#define SIGNS_MAX 20

/* Data association */
/* Distance type (you must chose only one) */
//#define NEAREST_NEIGHBOR_SQUARE_EUCLIDIAN_DISTANCE
#define NEAREST_NEIGHBOR_MAHALANOBIS_DISTANCE
#define DISTANCE_MAX 6.0 // threshold to associate a detection with a prediction

/* Detection and measurement */
#define DETECTIONS_MAX 128 // must be greater than SIGNS_MAX
#define MISSING_MEASURES_MAX 3 // threshold to stop tracking
#define AGE_TO_BE_A_SIGN FRAMES_MAX // minimum age to be considered as a real sign

/* Haar */
/* Don't forget to change conversion type (grayscale or normalized
 * red) in videoGrab() according to the cascade below. */
//#define HAAR_CASCADE_PATH "classifiers/Haar_red_triangle_up_Y_ls798_32x32"
//#define HAAR_CASCADE_PATH "classifiers/Haar_red_triangle_up_RedNorm_ls798_32x32"
#define HAAR_CASCADE_PATH "classifiers/Haar_cropRect_red_triangle_up_RedNorm_ls798_37x32"
//#define HAAR_CASCADE_PATH "classifiers/Haar_cropRect_red_triangle_up_Y_ls798_37x32_BASIC"
#define HAAR_IMAGE_SIZE cvSize(37,32)
#define HAAR_IMAGE_SIZE_MINIMUM_DETECTION cvSize(37,32)
#define HAAR_SCALE 1.1
#define HAAR_NEIGHBORS 2

/* -------------------------------------------------------------------------- */

/* Measure of a tracked sign. */
typedef struct measure_t
{
    int exist; // != 0 only if it's a real associated measure (not a
               // prediction)
    double confidence; // between 0 and 1
    CvMat* vector;
} Measure;

/* Subimage of a tracked sign. */
typedef struct patch_t
{
    int exist; // flag
    double confidence; // between 0 and 1
    /* ROI that contains the sign. */
    CvPoint topLeftPoint, bottomRightPoint;
} Patch;

/* A tracked sign. */
typedef struct sign_t
{
    CvKalman* kalman;
    int age; // number of frames for which the sign is tracked
    int numberOfMissingMeasures; // *successive* missing measures
    Measure measures[FRAMES_MAX];
    int numberOfPatches;
    Patch patches[FRAMES_MAX]; // subimages of the sign in frames
    int numberOfPredictions;
    int numberOfCorrections;
} Sign;

/* All tracked objects. */
typedef struct tracking_t
{
    int numberOfSigns; // number of actually tracked signs
    Sign signs[SIGNS_MAX];
    /* Detections from the cascade of Haar classifiers. */
    CvMemStorage* haarDetectionStorage;
    CvSeq* haarDetections;
    CvHaarClassifierCascade* cascade; // Haar cascade 
    /* Reformated detections sequence. */
    CvMemStorage* detectionStorage;
    CvSeq* detections;
    /* Each detection in the sequence should have its own detection
     * noise covariance but it's the same for all at present. */
    CvMat* detectionNoiseCov;
} Tracking;

/* Elements of detection sequence. */
typedef struct detection_t
{
    /* From Haar */
    CvRect rect;
    int neighbors;
    /* Others */
    CvPoint topLeftPoint, bottomRightPoint;
    CvPoint2D32f center;
} Detection;

/* -------------------------------------------------------------------------- */

/** Return 0 if a pixel is not red. */
inline int isRed( uchar* BGRPixel )
{
  uchar red = BGRPixel[2];
  //if( red != 0 && red > 1.1*BGRPixel[0] && red > 1.1*BGRPixel[1] )
  if( red != 0 && red >= BGRPixel[0] && red >= BGRPixel[1]
      && red > 1.15*BGRPixel[0] || red > 1.15*BGRPixel[1] )
      return 1;
  else
      return 0;
}

/* Initialize, allocate memory and free it. */
int initTracking( Tracking* tracking );
int freeTracking( Tracking* tracking );

/* Reset some values when we stop tracking that sign. */
void resetSign( Sign* sign );

/* Set transition matrix of the dynamix model. dt is in seconds. */
void setTransitionMatrix( Sign* sign, float dt );

/* Set process noise covariance matrix. */
void setProcessNoiseCov( Sign* sign );

/* Set measurement noise covariance matrix. */
void setMeasurementNoiseCov( Tracking* tracking, Sign* sign, double scale = 1 );

/* Set detection noise covariance matrix. At present, it's always the
 * same and for all signs. */
void setDetectionNoiseCov( Tracking* tracking );

/* Predict state of all tracked objects. */
void predictState( Tracking* tracking, Video* video );

/* Draw predictions and their error ellipses. */
void drawPredictions( Tracking* tracking, Video* video, IplImage* picture,
                      CvScalar color = CV_WHITE, int pointRadius = 1,
                      double ellipseScale = 1  );

/* This only a test function detecting 1-pixel-points to track
 * them. It's useful to check how Kalman filter works. */
void detectPoints( Tracking* tracking, Video* video );

/* Draw detections (from detectPoints()) and their error
 * ellipses. */
void drawDetectedPoints( Tracking* tracking, Video* video, IplImage* picture,
                         CvScalar color = CV_WHITE, int pointRadius = 1,
                         double ellipseScale = 1 );

/* Detect signs using a cascade of Haar classifiers on the current
 * frame. */
void detectSigns( Tracking* tracking, Video* video );

/* Eliminate patches from detectSigns() that don't seem to contain a
 * red sign. */
void keepGoodDetections( Tracking* tracking, Video* video );

/* Draw detections (from detectSigns()). */
void drawDetectedSigns( Tracking* tracking, Video* video, IplImage* picture,
                        CvScalar color = CV_WHITE, int thickness = 1,
                        int pointRadius = 1, double ellipseScale = 1 );

/* Draw Measures (detections which have been associated with a sign)
 * and their error ellipses. Indexes 2 are for signs older than a
 * threshold. */
void drawMeasures( Tracking* tracking, Video* video, IplImage* picture,
                   CvScalar color1 = CV_WHITE, int thickness1 = 1, int pointRadius1 = 1,
                   CvScalar color2 = CV_WHITE, int thickness2 = 1, int pointRadius2 = 1,
                   double ellipseScale = 1 );

/* Draw labels (signs number) near tracked signs. */
void drawLabels( Tracking* tracking, Video* video, IplImage* picture );

/* Gating and data association between tracked objects and detections
 * in current frame. */
void dataAssociation( Tracking* tracking, Video* video );

/* Compute distances between tracked signs and detections. Function
 * allocates memory to distanceMatrix and puts the results in it.
 * Matrix dimensions is SIGNS_MAX x number of detections. If one
 * dimension is zero, distanceMatrix = NULL. If the i-th sign is not
 * tracked, all element in line i equal -1. */
void computeDistances( Tracking* tracking, CvMat** distanceMatrix );

/* Come to a decision for each sign according to its past and
 * association index: convert association index into a measure, stop
 * tracking it, new tracked sign, etc. */
void convertIndexToMeasure( Tracking* tracking, Video* video, int* associations );

/* Correct state of all tracked objects. */
void correctState( Tracking* tracking, Video* video );

/* Draw corrections and their error ellipses. */
void drawCorrections( Tracking* tracking, Video* video, IplImage* picture,
                      CvScalar color = CV_WHITE, int pointRadius = 1,
                      double ellipseScale = 1  );

/* New sign require Kalman filter initialization. CAREFUL! Must be
 * called after a new videoGrab() (in prediction block for instance),
 * otherwise there are problems with frame index (view source for
 * further information). */
void initKalman( Sign* sign, Video* video );

/* Test if an image patch contain an up-triangular red sign. Return 0
 * if there is no sign. */
int isRedSign( const IplImage* src, CvRect roi, Video* video );

/* Test the size of a ROI and return 1 if it could contain a sign or 0
 * otherwise. */
int isGoodSize( CvRect roi );

/* Save patches (subimages) of signs taking into account only tracked
 * signs that are many frames old. */
void savePatches( Tracking* tracking, Video* video );

/* Draw (even in previous frames) a rectangle around signs that are
 * many frames old. */
void drawPatches( Tracking* tracking, Video* video, CvScalar color = CV_WHITE,
                  int thickness = 1 );

#endif
