/* Traffic signs detection and tracking. */

/* The program could run MUCH faster if the code was optimized.*/

#include <stdio.h>
#include <cv.h>
#include <highgui.h>
#include <libgen.h>
#include "video.h"
#include "misc.h"
#include "tracking.h"
#include "events.h"
#include "verbose.h"
#include "timing.h"
#include "color.h"

/* If you want less terminal outputs and display no images, set
 "#define VERBOSE 0" in verbose.h and comment "#define SHOW_IMAGES" in
 tsr.c */

#define SHOW_IMAGES
#define SKIP_FRAME 3 /* grab one frame on SKIP_FRAME in the video file */
#define PREDICTION_RECT_THICKNESS 1 /* prediction */
#define PREDICTION_COLOR CV_CYAN
#define DETECTION_RECT_THICKNESS 1 /* Haar detection */
#define DETECTION_COLOR CV_RED
#define GOOD_DETECTION_RECT_THICKNESS 1 /* Haar detection with several tests */
#define GOOD_DETECTION_COLOR CV_MAGENTA
#define MEASURE_RECT_THICKNESS 1 /* associated detection */
#define MEASURE_COLOR CV_GREEN
#define CORRECTION_RECT_THICKNESS 1 /* correction */
#define CORRECTION_COLOR CV_BLUE
#define PATCH_RECT_THICKNESS 2 /* detection taken as signs */
#define PATCH_COLOR CV_YELLOW

int _verbose_functionCallDepth = 0;

int main(int argc, char** argv)
{
  /* Test new functions on a random matrix. */
/*   CvMat* m = cvCreateMat( 3, 4, CV_32FC1 ); */
/*   CvRNG rng = cvRNG(-1); */
/*   cvRandArr( &rng, m, CV_RAND_NORMAL, cvRealScalar(0), cvRealScalar(0.1) ); */
/*   printMatrix( m ); */
/*   permuteMatrix32f( m, 2, 1, 1 ); */
/*   printMatrix( m ); */
/*   cvReleaseMat( &m ); */
  
  FUNCTION_BEGIN;
  Video video;
  Tracking tracking;
  char message[512];

  initVideo( &video, argc, argv );
  initTracking( &tracking );

  /* Main loop */
  for(;;)
  {
    /* Grab a new image. */
    IplImage* currentFrame = videoGrab( &video, SKIP_FRAME );
    if( currentFrame == NULL )
    {
      PRINT_ERR("NULL frame");
      break;
    }

    /* Copy current frame to display things. */
    cvCopy( currentFrame, video.display[0] );
    
    /* State prediction for all tracked objects. */
    predictState( &tracking, &video );
    /* Detection on current frame. */
    detectSigns( &tracking, &video );
#ifdef SHOW_IMAGES
    drawDetectedSigns( &tracking, &video, video.display[0], DETECTION_COLOR,
                       DETECTION_RECT_THICKNESS, 2 , 2);
#endif
    int numberOfHaarDetections = tracking.detections->total;
    keepGoodDetections( &tracking, &video );
    /* Gating and data association. */
    dataAssociation( &tracking, &video );
    /* State correction for all tracked objects. */
    correctState( &tracking, &video );
    /* Save subimages containing signs. */
    savePatches( &tracking, &video );

#if VERBOSE >= 9
  printTracking( &tracking, NULL, video.fi, 1 );
#endif
    
#if VERBOSE >= 10
    /* Print Kalman structure of all tracked signs. */
    for( int i = 0; i < SIGNS_MAX; i++)
    {
      Sign* sign = &tracking.signs[i];
      if( sign->age > 0 )
      {
        sprintf( message, "sign %d", i );
        printKalman( sign->kalman, message );
      }
    }
#endif

#ifdef SHOW_IMAGES
    /* Others Drawing */
    drawPredictions( &tracking, &video, video.display[0], PREDICTION_COLOR, 3, 2 );
    drawDetectedSigns( &tracking, &video, video.display[0], GOOD_DETECTION_COLOR,
                       DETECTION_RECT_THICKNESS, 2, 2 );
    drawMeasures( &tracking, &video, video.display[0],
                  MEASURE_COLOR, MEASURE_RECT_THICKNESS, 2,
                  PATCH_COLOR, PATCH_RECT_THICKNESS, 2,
                  2 );
    drawCorrections( &tracking, &video, video.display[0], CORRECTION_COLOR, 1, 2 );
    drawLabels( &tracking, &video, video.display[0] );
    /* Shows and eventually saves image. */   
    sprintf( message, " %d Haar detections, %d remaining detections, %d tracks",
             numberOfHaarDetections, tracking.detections->total, tracking.numberOfSigns );
    putText( video.display[0], message, cvPoint(0, VIDEO_HEIGHT - 6),
             &video.fonts.plainSmall, CV_BLACK, CV_WHITE, 1 );
    saveVideo( &video, &video.save[0], video.display[0], SAVE_IMAGES );
    displayImage( "TSR", video.display[0] );
    /* Final image containing only patches (with a frame delay). */
    int patchIndex = INDEX_CIRC_BUFFER(video.fi-(AGE_TO_BE_A_SIGN-1), FRAMES_MAX);
    drawPatches( &tracking, &video, PATCH_COLOR, PATCH_RECT_THICKNESS );
    if( video.numberOfGrabs >= AGE_TO_BE_A_SIGN )
        saveVideo( &video, &video.save[1], video.frames[patchIndex], SAVE_IMAGES );
#endif
    
    /* Events */
#ifdef SHOW_IMAGES
    switch( (char) cvWaitKey(20) )
    {
        case KEY_EXIT:
            goto exit;
            break;
        case KEY_PAUSE:
            cvWaitKey();
            break;
        case KEY_SAVE_VIDEO_0:
            video.save[0].request ^= 1;
            video.save[1].request ^= 1;
            break;
        default:
            break;
    }
#else
    /* We cannot use cvWaitKey() if there is no displayed image. We
     * don't handle events but we just break if ENTER is pressed. */
    fd_set rfds;
    struct timeval tv;
    int retval;
    /* Watch stdin (fd 0) to see when it has input. */
    FD_ZERO(&rfds);
    FD_SET(0, &rfds);
    /* Wait up to five seconds. */
    tv.tv_sec = 0;
    tv.tv_usec = 1;
    retval = select(1, &rfds, NULL, NULL, &tv);
    /* Don't rely on the value of tv now! */
    if( retval != -1 && retval != 0 )
        break;
#endif
  }

  /* Program termination */
#ifdef SHOW_IMAGES 
  exit:
  /* Saving remaining patched frames. */
  if( video.numberOfGrabs >= AGE_TO_BE_A_SIGN )
      for( int j = 0; j < AGE_TO_BE_A_SIGN; j++ )
      {
        int patchIndex = INDEX_CIRC_BUFFER(video.fi-(AGE_TO_BE_A_SIGN-1)+j, FRAMES_MAX);
        saveVideo( &video, &video.save[1], video.frames[patchIndex], SAVE_IMAGES );
      }
  PRINT( "Press any key to free memory and exit." );
  cvWaitKey();
  cvDestroyWindow( "TSR" );
#endif
  freeVideo( &video );
  freeTracking( &tracking );
  PRINT( "program termination" );
  FUNCTION_END;
  return 0;
}
