#include <cv.h>
#include <stdio.h>
#include "verbose.h"
#include "tracking.h"

/* -------------------------------------------------------------------------- */

void printMatrix( const CvMat* m, const char* message, CvRect* rect )
{
  FUNCTION_BEGIN;
  if( message )
  {
    PRINT1("%s", message);
  }

  int type = cvGetElemType( m );
  struct 
  {
      int i1; // Zero-based line index top-left element 
      int j1; // Zero-based column index top-left element
      int i2; // Zero-based index bottom-right element 
      int j2; // Zero-based column index bottom-right element
  } sub;
  if( rect )
  {
    sub.i1 = rect->y;
    sub.j1 = rect->x;
    sub.i2 = rect->height + sub.i1 -1;
    sub.j2 = rect->width + sub.j1 -1;
  }
  else
  {
    sub.i1 = 0;
    sub.j1 = 0;
    sub.i2 = m->rows - 1;
    sub.j2 = m->cols - 1;
  }

  PRINT1( "rows = %d (display = %d-%d), cols = %d (display = %d-%d), type = %d, step = %d",
          m->rows, sub.i1, sub.i2, m->cols, sub.j1, sub.j2, type, m->step);
  
  for( int i = sub.i1 ; i <= sub.i2; i++ )
  {
    FUNCTION_DEPTH_TAB(stdout);
    for( int j = sub.j1; j <= sub.j2; j++ )
    {
      switch (type)
      {
          case CV_16UC1:
              printf( "%u\t", ((unsigned short*)(m->data.ptr + m->step*i))[j] );
              break;
          case CV_16SC1:
              printf( "%hd\t", ((short*)(m->data.ptr + m->step*i))[j] );
              break;
          case CV_32SC1:
              printf( "%d\t", ((int*)(m->data.ptr + m->step*i))[j] );
              break;
          case CV_32FC1:
              printf( "%.3e\t", ((float*)(m->data.ptr + m->step*i))[j] );
              break;
          case CV_64FC1:
              printf( "%.3e\t", ((double*)(m->data.ptr + m->step*i))[j] );
              break;
          default:
              PRINT_ERR( "element type (%d) not supported", type );
              return;
      }
    }
    printf( "\n" );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void printArray( void* data, int rows, int cols, int type,
                 const char* message, CvRect* rect )
{
  FUNCTION_BEGIN;
  CvMat m;
  if( rows > 0 && cols > 0 )
  {
    cvInitMatHeader( &m, rows, cols, type, data );
    printMatrix( &m, message );
  }
  else
  {
    PRINT( "%s", message );
    PRINT_ERR( "Array dimensions must be positive! rows = %d, cols = %d",
               rows, cols );
  }
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void printKalman( const CvKalman* kalman, const char* message)
{
  FUNCTION_BEGIN;
  if( message )
  {
    PRINT1("%s", message);
  }

  printMatrix( kalman->state_pre, "state_pre" );
  printMatrix( kalman->state_post, "state_post" );
  printMatrix( kalman->transition_matrix, "transition_matrix" );
  printMatrix( kalman->process_noise_cov, "process_noise_cov" );
  printMatrix( kalman->measurement_matrix, "measurement_matrix" );
  printMatrix( kalman->measurement_noise_cov, "measurement_noise_cov" );
  printMatrix( kalman->error_cov_pre, "error_cov_pre" );
  printMatrix( kalman->gain, "gain" );
  printMatrix( kalman->error_cov_post, "error_cov_post" );
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void printCaptureProperties( CvCapture* capture, const char* message)
{
  FUNCTION_BEGIN;
  if( message )
  {
    PRINT1("%s", message);
  }
  
  PRINT1("CV_CAP_PROP_POS_MSEC: %f",
         cvGetCaptureProperty(capture, CV_CAP_PROP_POS_MSEC));
  PRINT1("CV_CAP_PROP_POS_FRAMES: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES));
  PRINT1("CV_CAP_PROP_POS_AVI_RATIO: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_POS_AVI_RATIO));
  PRINT1("CV_CAP_PROP_FRAME_WIDTH: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH));
  PRINT1("CV_CAP_PROP_FRAME_HEIGHT: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT));
  PRINT1("CV_CAP_PROP_FPS: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_FPS));
  PRINT1("CV_CAP_PROP_FOURCC: %f",
         cvGetCaptureProperty( capture, CV_CAP_PROP_FOURCC));
  FUNCTION_END;
}

/* -------------------------------------------------------------------------- */

void printTracking( Tracking* tracking, const char* message,
                    int frameIndex, int signAge )
{
  FUNCTION_BEGIN;
  if( message )
  {
    PRINT1( "%s", message );
  }

  PRINT1( "---------- numberOfSigns (tracked signs): %d ----------",
          tracking->numberOfSigns);
  for( int i = 0; i < SIGNS_MAX; i++ )
  {
    Sign* sign = &tracking->signs[i];
    if( sign->age >= signAge )
    {
      PRINT1( "----- sign %d -----", i);
      PRINT1( "age: %d", sign->age);
      PRINT1( "numberOfPatches: %d", sign->numberOfPatches);
      PRINT1( "numberOfMissingMeasures: %d", sign->numberOfMissingMeasures );
      PRINT1( "numberOfPredictions: %d", sign->numberOfPredictions );
      PRINT1( "numberOfCorrections: %d", sign->numberOfCorrections );
      if( frameIndex != -1 )
      {
        Measure* measure = &sign->measures[frameIndex];
        if( measure->exist != 0 )
        {
          printMatrix( measure->vector, "measure for this sign" );
        }
        else
        {
          PRINT1( "no measure for thi sign in current frame" );
        }
      }
    }
  }
  FUNCTION_END;
}
