#ifndef _VERBOSE_H
#define _VERBOSE_H

#include <cv.h>
#include <highgui.h>
#include <stdio.h>
#include <stdarg.h>
#include "tracking.h"
#include "color.h"

/* Verbosity Level */
#define VERBOSE 9

/* CAREFUL! When you use these macros, always put a ';' at end and
 * always put in block in a loop or "if". e.g.: if(a==4){MACRO;}. */


/* Fallback definition for older versions of GCC. View
 * http://gcc.gnu.org/onlinedocs/gcc/Function-Names.html */
#if __STDC_VERSION__ < 199901L
# if __GNUC__ >= 2
#  define __func__ __FUNCTION__
# else
#  define __func__ "<unknown>"
# endif
#endif


/* Enhanced printf as macros. It's variadic macros, view
 * http://fr.wikibooks.org/wiki/Programmation_C_Pr%C3%A9processeur or
 * http://www.redhat.com/docs/manuals/enterprise/RHEL-3-Manual/cpp/variadic-macros.html */
#define PRINT_STDOUT(message, ...) fprintf( stdout, "%s(): " message "\n", \
                                            __func__, ##__VA_ARGS__ )
#define PRINT_STDOUT_FULL(message, ...) fprintf( stdout, "%s():%s:%d: " message "\n", \
                                                 __func__, __FILE__, __LINE__, ##__VA_ARGS__ )
#define PRINT_STDERR(message, ...) fprintf( stderr, "%s(): " message "\n", \
                                            __func__, ##__VA_ARGS__ )
#define PRINT_STDERR_COLOR(message, ...) fprintf( stderr, "%s", TERM_RED ); \
                                         fprintf( stderr, "%s(): " message "\n", \
                                                  __func__, ##__VA_ARGS__ ); \
                                         fprintf( stderr, "%s", TERM_RESET )
#define PRINT_STDERR_FULL(message, ...) fprintf( stderr, "%s():%s:%d: " message "\n", \
                                                 __func__, __FILE__, __LINE__, ##__VA_ARGS__ )


/* We use macros FUNCTION_BEGIN and FUNCTION_END at begin and end of
 * all functions. Thus we know in which function the program has
 * crashed. */
extern int _verbose_functionCallDepth;

#define FUNCTION_DEPTH_TAB_ENABLE 1
#if FUNCTION_DEPTH_TAB_ENABLE
# define FUNCTION_DEPTH_TAB(fd) for( int _verbose_i = 0; \
                                     _verbose_i < 2*_verbose_functionCallDepth; \
                                     _verbose_i++ ) \
                                fprintf( fd, " ")
#else
# define FUNCTION_DEPTH_TAB(fd)
#endif

#define FUNCTION_BEGIN_END_ENABLE 1
#if FUNCTION_BEGIN_END_ENABLE && VERBOSE
# define FUNCTION_BEGIN FUNCTION_DEPTH_TAB(stdout); \
                        fprintf( stdout, "%s%s()%s: begin (%s%d%s)\n", \
                                 TERM_REVERS,__func__, TERM_RESET, TERM_GREEN, \
                                 _verbose_functionCallDepth, TERM_RESET ); \
                        _verbose_functionCallDepth++
# define FUNCTION_END _verbose_functionCallDepth--; \
                      FUNCTION_DEPTH_TAB(stdout); \
                      fprintf( stdout, "%s%s()%s: end   (%s%d%s)\n", \
                               TERM_REVERS,__func__, TERM_RESET, \
                               TERM_RED, _verbose_functionCallDepth, TERM_RESET )
#else
# define FUNCTION_BEGIN
# define FUNCTION_END
#endif


/* We define 10 verbosity levels (0-9). Level 0 correspond to no
 * output at all (except errors) and level 9 to full output including
 * debugging information. */
#if VERBOSE >= 1
# define PRINT1 FUNCTION_DEPTH_TAB(stdout); PRINT_STDOUT
# define PRINT_LINE FUNCTION_DEPTH_TAB(stdout); \
                    PRINT_STDOUT( "----------------------------------------" \
                                  "----------------------------------------" )
#else
# define PRINT1(...)
#endif

#if VERBOSE >= 9
# define PRINT9 FUNCTION_DEPTH_TAB(stdout); PRINT_STDOUT // for debugging information
#else
# define PRINT9(...)
#endif

#define PRINT_ERR FUNCTION_DEPTH_TAB(stderr); PRINT_STDERR_COLOR // for errors only
#define PRINT FUNCTION_DEPTH_TAB(stdout); PRINT_STDOUT // for anything

/* -------------------------------------------------------------------------- */

/* Print whole matrix on stdout. If rect is not NULL, print
 * correspondant sub-matrix. */
void printMatrix( const CvMat* m, const char* message = NULL, CvRect* rect = NULL);

/* Print array. If rect is not NULL, print correspondant
 * sub-array. e.g.:
 * int a[] = {1, 2, 3, 4, 5, 6};
 * printArray( (void*)a, 3, 2, CV_32SC1 ); */
void printArray( void* data, int rows, int cols, int type,
                 const char* message = NULL, CvRect* rect = NULL );

/* Print matrices of Kalman structure on stdout. */
void printKalman( const CvKalman* kalman, const char* message = NULL);

/* Print all video capturing properties on stdout. */
void printCaptureProperties( CvCapture* capture, const char* message = NULL );

/* Print Tracking structure on stdout. frameIndex is required if you
 * want to print measure in correspondant frame, otherwise no measure
 * is printed. Only signs older than signAge (or equal) are
 * printed. */
void printTracking( Tracking* tracking, const char* message = NULL,
                    int frameIndex = -1, int signAge = 0 );

#endif
