#include <highgui.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "video.h"
#include "verbose.h"
#include "timing.h"
#include "misc.h"

/* -------------------------------------------------------------------------- */

int initVideo( Video* video, int argc, char** argv )
{
  FUNCTION_BEGIN;
  /* Capture init */
  /* Default video file */
  if( argc == 1 )
  {
    video->capture = cvCaptureFromAVI( DEFAULT_VIDEO_FILE );
    video->sourceType = VIDEO_FILE_SOURCE;
    char* inputFileName;
    splitPath( DEFAULT_VIDEO_FILE, NULL, &inputFileName, NULL);
    sprintf( video->inputFileName, "%s", inputFileName );
    free(inputFileName);
    video->dts = video->fileDts;
  }
  else
  {
    char* inputFileName;
    splitPath( argv[1], NULL, &inputFileName, NULL);
    sprintf( video->inputFileName, "%s", inputFileName );
    free(inputFileName);
    /* Camera */
    if( argc == 2 && strlen(argv[1]) == 1 && isdigit(argv[1][0]) )
    {
      video->capture = cvCaptureFromCAM( argv[1][0] - '0' );
      video->sourceType = VIDEO_CAMERA_SOURCE;
      video->dts = video->realDts;
    }
    /* File */
    else if( argc == 2 )
    {
      video->capture = cvCaptureFromAVI( argv[1] );
      video->sourceType = VIDEO_FILE_SOURCE;
      video->dts = video->fileDts;
    }
  }

  if( video->capture == NULL )
  {
    PRINT_ERR("could not initialize capturing");
    exit(-1);
  }

  video->numberOfGrabs = 0;
  video->fi = 0;
  video->fps = 0;
  video->ti = 0;

  CvSize videoSize = cvSize(VIDEO_WIDTH, VIDEO_HEIGHT);
  /* Allocate images. */
  for( int i = 0; i < FRAMES_MAX; i++)
  {
    video->frames[i] = cvCreateImage( videoSize, 8, 3 );
    video->grayscaleFrames[i] = cvCreateImage( videoSize, 8, 1 );
  }
  for( int i = 0; i < FRAMES_MAX; i++)
  {
    video->frames[i] = cvCreateImage( videoSize, 8, 3 );
    video->grayscaleFrames[i] = cvCreateImage( videoSize, 8, 1 );
  }
  for( int i = 0; i < DISPLAYIMAGES_MAX; i++)
  {
    video->display[i] = cvCreateImage( videoSize, 8, 3 );
  }
  video->mask = cvCreateImage( videoSize, 8, 1 );

  /* Video saving*/
  for( int i = 0; i < SAVEVIDEOS_MAX; i++ )
  {
    SaveVideo* save = &video->save[i];
    save->writer = NULL;
    save->request = 0;
    save->on = 0;
  }

  /* Fonts */
  cvInitFont( &video->fonts.plain, CV_FONT_HERSHEY_PLAIN , 1, 1 );
  cvInitFont( &video->fonts.plainBold, CV_FONT_HERSHEY_PLAIN , 1, 1, 0, 2 );
  cvInitFont( &video->fonts.plainSmall, CV_FONT_HERSHEY_PLAIN , 0.8, 0.8 );
  cvInitFont( &video->fonts.plainSmallBold, CV_FONT_HERSHEY_PLAIN , 0.8, 0.8, 0, 2 );

  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

int freeVideo( Video* video )
{
  FUNCTION_BEGIN;
  cvReleaseCapture( &video->capture );
  for( int i = 0; i < FRAMES_MAX; i++)
  {
    PRINT9("free images for frame %d", i);
    cvReleaseImage( &video->frames[i] );
    cvReleaseImage( &video->grayscaleFrames[i] );
  }
  PRINT9("free other images");
  for( int i = 0; i < DISPLAYIMAGES_MAX; i++)
  {
    cvReleaseImage( &video->display[i] );
  }
  cvReleaseImage( &video->mask );

  for( int i = 0; i < SAVEVIDEOS_MAX; i++ )
  {
    CvVideoWriter* writer = video->save[i].writer;
    if( writer != NULL )
        cvReleaseVideoWriter( &writer );
  }
  FUNCTION_END;
  return 0;
}

/* -------------------------------------------------------------------------- */

IplImage* getFrameFromFile( CvCapture* capture, int skipFrame,
                            IplImage* copiedFrame )
{
  FUNCTION_BEGIN;
  if( skipFrame == 0 )
  {
    PRINT_ERR("skipFrame must be > 0");
    return NULL;
  }

  /* Get the requested frame. */
  IplImage* frame = NULL;
#define CVSETCAPTUREPROPERTY_BUG 1
#if CVSETCAPTUREPROPERTY_BUG
  /* Hack because cvSetCaptureProperty() doesn't work. We can't jump
   * directly to a given frame, so we query each frame until it's the
   * one we want (time-consuming). */
  double newPosition;
  for( int i = 0; i < skipFrame; i++ )
  {
    newPosition = cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES );
    frame = cvQueryFrame( capture ); /* no need to free frame memory */
    if( newPosition == cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES ) )
    {
      PRINT9("end of stream");
      FUNCTION_END;
      return NULL;
    }
  }
#else
  double position = cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES ) - 1;
  double newPosition = cvSetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES ,
                                             position + skipFrame );  
  if( newPosition != position + skipFrame )
  {
    PRINT9("end of stream");
    return NULL;
  }
  frame = cvQueryFrame( capture ); /* no need to free frame memory */
#endif

  /* Copy and return frame. */
  if( copiedFrame )
  {
    cvConvertImage( frame, copiedFrame );
    FUNCTION_END;
    return copiedFrame;
  }
  else
  {
    FUNCTION_END;
    return frame;
  }
}

/* -------------------------------------------------------------------------- */

IplImage* getFrameFromCam( CvCapture* capture, IplImage* copiedFrame )
{
  FUNCTION_BEGIN;
  /* Get the requested frame. */
  IplImage* frame = NULL;
  frame = cvQueryFrame( capture ); /* no need to free frame memory */
  /* Copy and return frame. */
  if( copiedFrame )
  {
    cvConvertImage( frame, copiedFrame );
    FUNCTION_END;
    return copiedFrame;
  }
  else
  {
    FUNCTION_END;
    return frame;
  }
}

/* -------------------------------------------------------------------------- */

IplImage* videoGrab( Video* video, int skipFrame )
{
  FUNCTION_BEGIN;
  video->fi = video->numberOfGrabs % FRAMES_MAX;
  IplImage* frame = NULL;
  /* Get frame. */
  switch( video->sourceType )
  {
      case VIDEO_FILE_SOURCE:
          PRINT( "--------------- grab %d (frame %d in file) @ %f fps (real) --------------",
                 video->numberOfGrabs + 1,
                 (int) cvGetCaptureProperty( video->capture, CV_CAP_PROP_POS_FRAMES ),
                 video->fps );
          frame = getFrameFromFile( video->capture, skipFrame,
                                    video->frames[video->fi] );
          break;
      case VIDEO_CAMERA_SOURCE:
          PRINT( "--------------- grab %d from camera @ %f fps --------------",
                 video->numberOfGrabs + 1, video->fps );
          frame = getFrameFromCam( video->capture, video->frames[video->fi] );
          break;
      default:
          PRINT_ERR("invalid source type");
          break;
  }
  
  if( frame == NULL )
  {
    PRINT_ERR("NULL frame");
    FUNCTION_END;
    return frame;
  }

  /* Others images */
  /* Conversion type (grayscale or normalized red) depends of the
   * detector choosed in tracking.h. */
  //cvConvertImage( frame, video->grayscaleFrames[video->fi] );
  cvtRedNorm( frame, video->grayscaleFrames[video->fi] );

  /* Real times */
  int ti = video->numberOfGrabs % TIMES_MAX;
  video->ti = ti;
  video->times[ti] = cvGetTickCount();
  static double cpuFreqHz = cvGetTickFrequency() * 1e6;
  double dt = video->times[ti] - video->times[(ti-1) % TIMES_MAX];
  video->dts[ti] = dt / cpuFreqHz;
  video->numberOfGrabs++;
  if( video->numberOfGrabs >= TIMES_MAX )
  {
    dt = video->times[ti] - video->times[(ti+1) % TIMES_MAX];
    video->fps = TIMES_MAX / (dt / cpuFreqHz);
  }

  /* File times */
  video->fileDts[ti] = (double) skipFrame / VIDEO_FPS;
  
  FUNCTION_END;
  return frame;
}

/* -------------------------------------------------------------------------- */

void saveVideo( Video* video, SaveVideo* save, IplImage* frame,
                int saveFlags )
{
  FUNCTION_BEGIN;
  if( save->request != 0 )
  {
    int i = 0;
    switch( save->on )
    {
        case 0:
            /* Search for a directory that doesn't exist. */
            char dir[512];
            while(1)
            {
              sprintf( dir, "%s-%02d", video->inputFileName, i );
              FILE* fd = fopen( dir, "r" );
              if( fd == NULL )
                  break;
              else
                  fclose( fd );
              i++;
            }
            sprintf( save->dir, "%s", dir );
            mkdir( save->dir, 0777 );
            char optionsFileName[512];
            sprintf( optionsFileName, "%s/options.txt", dir );
            saveOptions( optionsFileName );
            if( saveFlags & SAVE_VIDEO )
            {
              sprintf( save->videoName, "%s/video.mpg", dir );
              /* opencv-grep-sources.sh MKTAG to see some FOURRCC. We
               * have to choose a fps consistent with MPEG codec, we
               * choose 25. To replay video you can use "mplayer -fps 1
               * videoname". */
              PRINT9("creating %s", save->videoName );
              save->writer = cvCreateVideoWriter( save->videoName,
                                                  //CV_FOURCC('D', 'I', 'V', 'X'),
                                                  //CV_FOURCC('D', 'I', 'V', '3'),
                                                  //CV_FOURCC('M', 'P', '4', '2'),
                                                  //CV_FOURCC('M', 'J', 'P', 'G'),
                                                  //CV_FOURCC('U', '2', '6', '3'),
                                                  CV_FOURCC('P', 'I', 'M', '1'),
                                                  25, cvSize(VIDEO_WIDTH, VIDEO_HEIGHT));
            }
            save->on = 1;
            save->imageCounter = 0;
        default:
            if( saveFlags & SAVE_VIDEO )
            {
              PRINT9( "writing frame in %s", save->videoName );
              cvWriteFrame( save->writer, frame );
            }
            if( saveFlags & SAVE_IMAGES )
            {
              PRINT9( "writing frame in %s", save->imageName );
              sprintf( save->imageName, "%s/%06d.png",
                       save->dir, save->imageCounter );
              cvSaveImage( save->imageName, frame );
              save->imageCounter++;
            }
            putText( frame, "REC", cvPoint(1,10), &video->fonts.plainSmall,
                     CV_WHITE, CV_RED );
            break;
    }
  }
  else
  {
    switch( save->on )
    {
        case 1:
            PRINT9("closing %s", save->videoName );
            cvReleaseVideoWriter( &save->writer );
            save->on = 0;
        default:
            break;
    }
  }
  FUNCTION_END;
}
